#include "sdk_config.h"
#include "zboss_api.h"
#include "zboss_api_addons.h"
#include "zb_mem_config_med.h"
#include "zb_error_handler.h"
#include "zigbee_helpers.h"

#include "app_timer.h"
#include "bsp.h"
#include "boards.h"
#include "sensorsim.h"

#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

#include "zb_zcl_custom_thermostat.h"
#include "zb_custom_ha_thermostat.h"

#define IEEE_CHANNEL_MASK                   (1l << ZIGBEE_CHANNEL)              /**< Scan only one, predefined channel to find the coordinator. */
#define ERASE_PERSISTENT_CONFIG             ZB_TRUE                            /**< Erase NVRAM to save the network parameters after device reboot or power-off. NOTE: If this option is set to ZB_TRUE then do full device erase for all network devices before running other samples. */

#define HA_THERMOSTAT_ENDPOINT              9                                  /**< Device endpoint */

#define ZIGBEE_NETWORK_STATE_LED            BSP_BOARD_LED_2                      /**< LED indicating that light switch successfully joind Zigbee network. */

#define MIN_LOCAL_TEMPERATURE_VALUE         0                                    /**< Minimum local temperature value as returned by the simulated measurement function. */
#define MAX_LOCAL_TEMPERATURE_VALUE         4000                                 /**< Maximum local temperature value as returned by the simulated measurement function. */
#define LOCAL_TEMPERATURE_VALUE_INCREMENT   50                                   /**< Value by which the temperature value is incremented/decremented for each call to the simulated measurement function. */

#if !defined ZB_ED_ROLE
#error Define ZB_ED_ROLE to compile End Device source code.
#endif

static custom_ha_thermostat_device_ctx_t m_dev_ctx;

/* Declare attribute list for Basic cluster. */
ZB_ZCL_DECLARE_BASIC_ATTRIB_LIST(basic_attr_list,
                                 &m_dev_ctx.basic_attr.zcl_version,
                                 &m_dev_ctx.basic_attr.power_source);

/* Declare attribute list for Identify cluster. */
ZB_ZCL_DECLARE_IDENTIFY_ATTRIB_LIST(identify_attr_list,
                                    &m_dev_ctx.identify_attr.identify_time);

/* Declare attribute list for Thermostat cluster. */
ZB_ZCL_DECLARE_THERMOSTAT_ATTRIB_LIST(thermostat_attr_list,
                                      &m_dev_ctx.thermostat_attr.local_temperature,
                                      &m_dev_ctx.thermostat_attr.local_temperature_calibration,
                                      &m_dev_ctx.thermostat_attr.occupied_cooling_setpoint,
                                      &m_dev_ctx.thermostat_attr.occupied_heating_setpoint,
                                      &m_dev_ctx.thermostat_attr.control_seq_of_operation,
                                      &m_dev_ctx.thermostat_attr.system_mode,
                                      &m_dev_ctx.thermostat_attr.start_of_week);

/* Declare attribute list for Fan Control cluster. */
ZB_ZCL_DECLARE_FAN_CONTROL_ATTRIB_LIST(fan_control_attr_list,
                                       &m_dev_ctx.fan_mode_attr.fan_mode,
                                       &m_dev_ctx.fan_mode_attr.fan_mode_sequence);

/* Declare attribute list for Thermostat UI Configuration cluster. */
ZB_ZCL_DECLARE_THERMOSTAT_UI_CONFIG_ATTRIB_LIST(thermostat_ui_config_attr_list,
                                                &m_dev_ctx.thermostat_ui_config_attr.temperature_display_mode,
                                                &m_dev_ctx.thermostat_ui_config_attr.keypad_lockout);

ZB_ZCL_DECLARE_CUSTOM_THERMOSTAT_ATTRIB_LIST(custom_thermostat_attr_list,
                                            &m_dev_ctx.custom_thermostat_attr.thermostat_valve);

ZB_HA_DECLARE_CUSTOM_THERMOSTAT_CLUSTER_LIST(custom_thermostat_clusters,
                                  basic_attr_list,
                                  identify_attr_list,
                                  thermostat_attr_list,
                                  fan_control_attr_list,
                                  thermostat_ui_config_attr_list,
                                  custom_thermostat_attr_list);

/* Declare endpoint for Thermostat device. */
ZB_HA_DECLARE_CUSTOM_THERMOSTAT_EP(custom_thermostat_ep,
                            HA_THERMOSTAT_ENDPOINT,
                            custom_thermostat_clusters);

ZB_HA_DECLARE_CUSTOM_THERMOSTAT_CTX(thermostat_ctx,
                                    custom_thermostat_ep);

static sensorsim_cfg_t   m_local_temperature_sim_cfg;                                 /**< Temperature sensor simulator configuration. */
static sensorsim_state_t m_local_temperature_sim_state;                               /**< Temperature sensor simulator state. */

APP_TIMER_DEF(zb_app_timer);

/**@brief Function for initializing the application timer.
 */
static void timer_init(void)
{
    uint32_t error_code = app_timer_init();
    APP_ERROR_CHECK(error_code);
}

/**@brief Function for initializing the nrf log module.
 */
static void log_init(void)
{
    ret_code_t err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_DEFAULT_BACKENDS_INIT();
}

/**@brief Function for initializing LEDs.
 */
static zb_void_t leds_init(void)
{
    ret_code_t error_code;

    /* Initialize LEDs and buttons - use BSP to control them. */
    error_code = bsp_init(BSP_INIT_LEDS, NULL);
    APP_ERROR_CHECK(error_code);

    bsp_board_leds_off();
}

/**@brief Function for initializing all clusters attributes.
 */
static void custom_ha_thermostat_clusters_attr_init(void)
{
    /* Basic cluster attributes data */
    m_dev_ctx.basic_attr.zcl_version  = ZB_ZCL_VERSION;
    m_dev_ctx.basic_attr.power_source = ZB_ZCL_BASIC_POWER_SOURCE_UNKNOWN;

    /* Identify cluster attributes data */
    m_dev_ctx.identify_attr.identify_time        = ZB_ZCL_IDENTIFY_IDENTIFY_TIME_DEFAULT_VALUE;

    /* Thermostat cluster attributes data */
    m_dev_ctx.thermostat_attr.local_temperature             = ZB_ZCL_THERMOSTAT_LOCAL_TEMPERATURE_INVALID;
    m_dev_ctx.thermostat_attr.local_temperature_calibration = ZB_ZCL_THERMOSTAT_LOCAL_TEMPERATURE_CALIBRATION_DEFAULT_VALUE;
    m_dev_ctx.thermostat_attr.occupied_cooling_setpoint     = ZB_ZCL_THERMOSTAT_OCCUPIED_COOLING_SETPOINT_DEFAULT_VALUE;
    m_dev_ctx.thermostat_attr.occupied_heating_setpoint     = ZB_ZCL_THERMOSTAT_OCCUPIED_HEATING_SETPOINT_DEFAULT_VALUE;
    m_dev_ctx.thermostat_attr.control_seq_of_operation      = ZB_ZCL_THERMOSTAT_CONTROL_SEQ_OF_OPERATION_DEFAULT_VALUE;
    m_dev_ctx.thermostat_attr.system_mode                   = ZB_ZCL_THERMOSTAT_CONTROL_SYSTEM_MODE_DEFAULT_VALUE;
    m_dev_ctx.thermostat_attr.start_of_week                 = ZB_ZCL_THERMOSTAT_START_OF_WEEK_DEFAULT_VALUE;

    /* Fan Control cluster attributes data */
    m_dev_ctx.fan_mode_attr.fan_mode          = ZB_ZCL_FAN_CONTROL_FAN_MODE_DEFAULT_VALUE;
    m_dev_ctx.fan_mode_attr.fan_mode_sequence = ZB_ZCL_FAN_CONTROL_FAN_MODE_SEQUENCE_DEFAULT_VALUE;

    /* Thermostat UI Configuration cluster attributes data */
    m_dev_ctx.thermostat_ui_config_attr.temperature_display_mode = ZB_ZCL_THERMOSTAT_UI_CONFIG_TEMPERATURE_DISPLAY_MODE_DEFAULT_VALUE;
    m_dev_ctx.thermostat_ui_config_attr.keypad_lockout           = ZB_ZCL_THERMOSTAT_UI_CONFIG_KEYPAD_LOCKOUT_DEFAULT_VALUE;

    /* Custom thermostat cluster attributes data */
    m_dev_ctx.custom_thermostat_attr.thermostat_valve = ZB_ZCL_THERMOSTAT_VALVE_DEFAULT_VALUE;
}

/**@brief Function for handling nrf app timer.
 * 
 * @param[IN]   context   Void pointer to context function is called with.
 * 
 * @details Function is called with pointer to sensor_device_ep_ctx_t as argument.
 */
static void zb_app_timer_handler(void * context)
{
    zb_zcl_status_t zcl_status;
    static zb_int16_t new_local_temp_value;
    static zb_int16_t new_thermostat_valve_value;

    /* Get new temperature measured value */
    new_local_temp_value = (zb_int16_t)sensorsim_measure(&m_local_temperature_sim_state, &m_local_temperature_sim_cfg);
    zcl_status = zb_zcl_set_attr_val(HA_THERMOSTAT_ENDPOINT, 
                                     ZB_ZCL_CLUSTER_ID_THERMOSTAT, 
                                     ZB_ZCL_CLUSTER_SERVER_ROLE, 
                                     ZB_ZCL_ATTR_THERMOSTAT_LOCAL_TEMPERATURE_ID, 
                                     (zb_uint8_t *)&new_local_temp_value, 
                                     ZB_FALSE);
    if(zcl_status != ZB_ZCL_STATUS_SUCCESS)
    {
        NRF_LOG_INFO("Set local temperature value fail. zcl_status: %d", zcl_status);
    }

    new_thermostat_valve_value = new_local_temp_value-10;
    zcl_status = zb_zcl_set_attr_val(HA_THERMOSTAT_ENDPOINT, 
                                     ZB_ZCL_CLUSTER_ID_CUSTOM_THERMOSTAT, 
                                     ZB_ZCL_CLUSTER_SERVER_ROLE, 
                                     ZB_ZCL_ATTR_THERMOSTAT_VALVE_ID, 
                                     (zb_uint8_t *)&new_thermostat_valve_value, 
                                     ZB_FALSE);
    if(zcl_status != ZB_ZCL_STATUS_SUCCESS)
    {
        NRF_LOG_INFO("Set thermostat valve value fail. zcl_status: %d", zcl_status);
    }
}

/**@brief Function for initializing the sensor simulator.
 */
static void sensor_simulator_init(void)
{
    m_local_temperature_sim_cfg.min          = MIN_LOCAL_TEMPERATURE_VALUE;
    m_local_temperature_sim_cfg.max          = MAX_LOCAL_TEMPERATURE_VALUE;
    m_local_temperature_sim_cfg.incr         = LOCAL_TEMPERATURE_VALUE_INCREMENT;
    m_local_temperature_sim_cfg.start_at_max = false;

    sensorsim_init(&m_local_temperature_sim_state, &m_local_temperature_sim_cfg);
}

/*
  This function does not do anything right now, as no ZCL commands are used in this example.
  The function is left here in case cluster commands shall be added.
*/

/**@brief Callback function for handling ZCL commands.
 *
 * @param[in]   bufid   Reference to Zigbee stack buffer used to pass received data.
 */
static zb_void_t zcl_device_cb(zb_bufid_t bufid)
{
    zb_uint8_t                       cluster_id;
    zb_uint8_t                       attr_id;
    zb_zcl_device_callback_param_t * p_device_cb_param = ZB_BUF_GET_PARAM(bufid, zb_zcl_device_callback_param_t);

    NRF_LOG_INFO("zcl_device_cb id %hd", p_device_cb_param->device_cb_id);

    /* Set default response value. */
    p_device_cb_param->status = RET_OK;

    NRF_LOG_INFO("zcl_device_cb status: %hd", p_device_cb_param->status);
}

/**@brief Zigbee stack event handler.
 *
 * @param[in]   bufid   Reference to the Zigbee stack buffer used to pass signal.
 */
void zboss_signal_handler(zb_bufid_t bufid)
{
    zb_zdo_app_signal_hdr_t  * p_sg_p      = NULL;
    zb_zdo_app_signal_type_t   sig         = zb_get_app_signal(bufid, &p_sg_p);
    zb_ret_t                   status      = ZB_GET_APP_SIGNAL_STATUS(bufid);

    /* Update network status LED */
    zigbee_led_status_update(bufid, ZIGBEE_NETWORK_STATE_LED);

    switch (sig)
    {
        case ZB_BDB_SIGNAL_DEVICE_REBOOT:
            /* fall-through */
        case ZB_BDB_SIGNAL_STEERING:
            /* Call default signal handler. */
            ZB_ERROR_CHECK(zigbee_default_signal_handler(bufid));
            if (status == RET_OK)
            {
                ret_code_t err_code = app_timer_start(zb_app_timer, APP_TIMER_TICKS(1000), NULL);
                APP_ERROR_CHECK(err_code);
            }
            break;

        default:
            /* Call default signal handler. */
            ZB_ERROR_CHECK(zigbee_default_signal_handler(bufid));
            break;
    }

    if (bufid)
    {
        zb_buf_free(bufid);
    }
}

/**@brief Function for application main entry.
 */
int main(void)
{
    zb_ret_t       zb_err_code;
    ret_code_t     err_code;
    zb_ieee_addr_t ieee_addr;
    
    /* Initialize logging system and GPIOs. */
    timer_init();
    log_init();
    sensor_simulator_init();
    leds_init();

    /* Create Timer for reporting attribute */
    err_code = app_timer_create(&zb_app_timer, APP_TIMER_MODE_REPEATED, zb_app_timer_handler);
    APP_ERROR_CHECK(err_code);

    /* Set Zigbee stack logging level and traffic dump subsystem. */
    ZB_SET_TRACE_LEVEL(ZIGBEE_TRACE_LEVEL);
    ZB_SET_TRACE_MASK(ZIGBEE_TRACE_MASK);
    ZB_SET_TRAF_DUMP_OFF();

    /* Initialize Zigbee stack. */
    ZB_INIT("thermostat");

    /* Set device address to the value read from FICR registers. */
    zb_osif_get_ieee_eui64(ieee_addr);
    zb_set_long_address(ieee_addr);

    /* Set static long IEEE address. */
    zb_set_network_ed_role(IEEE_CHANNEL_MASK);
    zigbee_erase_persistent_storage(ERASE_PERSISTENT_CONFIG);

    zb_set_ed_timeout(ED_AGING_TIMEOUT_64MIN);
    zb_set_keepalive_timeout(ZB_MILLISECONDS_TO_BEACON_INTERVAL(3000));
    zb_set_rx_on_when_idle(ZB_FALSE);

    /* Initialize application context structure. */
    UNUSED_RETURN_VALUE(ZB_MEMSET(&m_dev_ctx, 0, sizeof(m_dev_ctx)));

    /* Not actually used, since no ZCL commands in this example. */
    /* Register callback for handling ZCL commands. */
    ZB_ZCL_REGISTER_DEVICE_CB(zcl_device_cb);

    /* Register dimmer switch device context (endpoints). */
    ZB_AF_REGISTER_DEVICE_CTX(&thermostat_ctx);
    custom_ha_thermostat_clusters_attr_init();

    /** Start Zigbee Stack. */
    zb_err_code = zboss_start_no_autostart();
    ZB_ERROR_CHECK(zb_err_code);

    while(1)
    {
        zboss_main_loop_iteration();
        UNUSED_RETURN_VALUE(NRF_LOG_PROCESS());
    }
}